'''
## This file is autogenerated from macrosynergy setup.py.

It contains version information for the macrosynergy package.

short_version: 1.5.4

version: 1.5.4

full_version: 1.5.4

git_revision: dded58ac486fe0bbc2580d5ecba089b6c4c2a8e8

release: True

'''
short_version: str = '1.5.4'
version: str = '1.5.4'
full_version: str = '1.5.4'
git_revision: str = 'dded58ac486fe0bbc2580d5ecba089b6c4c2a8e8'
release: bool = True
if not release:
    version = full_version
